package org.jum.scalademo

object ParameterizedTypeExample {
  
     case class A();
     case class B() extends A;
     
     def main(args : Array[String]) : Unit = {
       println("Variances");
       


     
     var a : A = new A();
     var b : B = new B();
    
    
    //Egyszerű osztály típusparaméter nélkül.
    class Simple(val value : A);
    val simple : Simple = new Simple(new A());
    a = simple.value;
    
    
    
    
    
    //Típusparaméter alkalmazása.
    class InvariantGeneric[T](val value : T);
    
    
    
    
    
    
    //Ugyanolyan típusú értéket teszünk be, mint amit kiolvasunk.
    val invariantGeneric1 : InvariantGeneric[A] =
      new InvariantGeneric[A](new A());
    a = invariantGeneric1.value;
    
    
    
    //A típusparaméter leszármazottját tesszük be,
    //de akkor is a típusparaméternek
    //megfelelő típust olvassuk vissza.
    val invariantGeneric2 : InvariantGeneric[A] =
      new InvariantGeneric[A](new B());
    a = invariantGeneric2.value;
    //b = invariantGeneric2.value; Type mismatch; found: A, required: B. 
    
    //Az alábbi nem fog lefordulni a generikus szerkezetek invarianciája miatt, azaz
    //InvariantGeneric[A] típusú elemnek csak InvariantGeneric[A] típusú értéket lehet adni.
    //val invariantGeneric3 : InvariantGeneric[A] = new InvariantGeneric[B](new B());
      
    
      
      
      
    //Meg lehet adni, hogy a generikus osztály kovariánsan viselkedjen, azaz ha B típusparaméter
    //leszármazottja A típusparaméternek, akkor CovariantGeneric[B] leszármazottja legyen CovariantGeneric[A]-nek.
    class CovariantGeneric[+T](val value: T);

    
    val covariantGeneric1 : CovariantGeneric[A] =
      new CovariantGeneric[B](new B());
    a = covariantGeneric1.value;
    //Az alábbi nem fog lefordulni, mert hiába CovariantGeneric[B] a futási idejű típus,
    //CovariantGeneric[A]-ként van deklarálva az elem.
    //b = covariantGeneric1.value;
      
    //Az automatikus dedukció miatt nem muszáj kiírni a konstruktorba a típusparamétert,
    //viszont akkor az elem típusparaméterére illeszti rá.
    val covariantGeneric2 : CovariantGeneric[A] =
      new CovariantGeneric(new B());
    //Tehát akkor a következő értékadás érvénytelen:
    //b = covariantGeneric2.value;
      
          
    //Java: <? extends T>
    
    //Nem fordul: csak immutable szerkezetekre alkalmazható.
    //class CovariantGeneric2[+T](var value: T);
    
    //Például a scala.collection.Map egyik típusparamétere is kovariánsként van deklarálva: Map[K, +V]
    val m1 : Map[String, A] = Map[String, B]("b" -> new B());
    a = m1("b");
    println(a);
    
    
    //Kontravariancia: Java: <? super T>
    
    
    
    
    
    //Scala függvényeknél: Function[-P1, -P2, ... +R]
    //Elvont példa függvény varianciára:
   
    def hof(e : (B, B) => A) = {};
    
    
    def f1(p1: B, p2 : B) : A = new A;
    hof(f1);
    
    
    def f2(p1: A, p2: A) : B = new B;
    
    //f1 paramétereit mindig elfogadja f2 is, mivel B extends A.
    //f2 visszatérési értéke kompatibilis f1 visszatérési értékével.
    //f1 helyére behelyettesíthető f2:
    
    hof(f2);
  }
     
     class F1 {
       def f(a: A) : A = new A;
     }
     
     class F2 extends F1 {
       override def f(a: A) : B = new B;
     }
}
