package org.jum.scalademo

//Example is taken from ProgrammingScalaBook online version.


//Some graphical widget
abstract class Widget





//A simple button
class Button(val label: String) extends Widget {
  def click() = {
    // Logic to give the appearance of clicking a button...
    println("Button is clicked.")
  }
}





//Observer behaviour
trait Observable {
  type Observer = { def receiveUpdate(subject: Any) }

  private var observers = List[Observer]()
  def addObserver(observer: Observer) = observers ::= observer
  protected def notifyObservers = observers foreach (_.receiveUpdate(this))
}




//Simple class with trait
class ObservableButton(name: String) extends
  Button(name) with Observable {
  override def click() = {
    super.click()
    notifyObservers
  }
}


class ButtonCountObserver {
  var count = 0
  def receiveUpdate(subject: Any) = count += 1
}


//See it in action
object TraitExample {
  def main(args:Array[String]) : Unit = {
    
      val observableButton1 = new ObservableButton("Okay")
      val buttonObserver1 = new ButtonCountObserver
      observableButton1.addObserver(buttonObserver1)
      
      for (i <- 1 to 3) observableButton1.click()
      println("Click count: " + buttonObserver1.count);
      
      
      
      //Adding trait at instaintiation time
      val observableButton2 = new Button("Okay")
with Observable {
          override def click() = {
        	  super.click()
        	  notifyObservers
          }
      }
      val buttonObserver2 = new ButtonCountObserver
      observableButton2.addObserver(buttonObserver2)

      for (i <- 1 to 3) observableButton2.click()
      println("Click count: " + buttonObserver2.count);
  }
}